///////////////////////////////////////////////////////////////////
// J-ACE (Jaka Arcade Controls Encoder) v0.8 for HW v.1.0
// Copyright 2020 Jaakko Kairus
// jaakko@kair.us
//
// You may freely use and modify this software for personal use.
// Commercial use of this software or parts of it is prohibited
// without written permission from author.
// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY
//
// For more information about this project see project web page at
// http://kair.us/projects/j_ace/
//
// Compiled with CCS compiler version 5.097
// Total compiled length 4702 words for keyboard mode
// Total compiled length 4052 words for joystick mode
// Program lengths include 1024 word bootloader
///////////////////////////////////////////////////////////////////
// Recent changes:
// v0.1 13.2.2020
// Initial version. Enumerates as 3 HID boot protocol keyboards
// Keys dedicated to a certain keyboard. Length 3486 words.
// v0.2 15.2.2020
// Now all keys are prioritised to first keyboard. 2nd and 3rd
// are used if all keys don't fit to first report. Length 3478 words.
// v0.3 19.2.2020
// Added switch debounce function. Length 3660 words.
// v0.4 19.2.2020
// Added selectable keymaps for players 1,2,3 or 4.
// Added 'shift' function for player 1 keymap, start button works as shift
// and changes keycodes of joystick and action buttons
// Length 3784 words
// v0.5 30.3.2020
// Changed debouncne function to use timer so that it
// is independent of USB polling rate. 3959 words
// 420 us for button readout and debounce
// 350 us for just the debounce
// Verified that actual polling interval is still 1ms even if all 3
// keyboards are sending data.
// Should add a way to either rescan buttons until USB polls data
// or wait for known time before reading buttons so that USB poll
// will happen soon after it, to minimize delay especially if poll
// rate is less than 1000 Hz
// Added autofire. 
// Autofire has configurable speed and buttons. Autofire LED
// indicates speed, activation and when in configure mode.
// Autofire speed adjusts with polling rare. Length 4109 words.
// 420us total for usb_keyboard_task() before sending data
// 500us when global autofire is on
// each autofire button down adds ca. 10 us.
// Added saving of autofire speed and autofire buttons.
// Added player/keymap setup mode (keep button 1 pressed when
// plugging in USB cable). Player/keymap also saved.
// Checking validity of settings saved to EEPROM. Length 4375
// words.
// v0.6 1.4.2020
// Splitted usb_keyboard_task() into several separate functions
// Allows to enable/disable different functions easily, makes
// code easier to read and prepares to add also joystick mode
// in addition to keyboard encoder mode. Length 4387 words.
// v0.7 5.4.2020
// Added joystick mode. Define J_ACE_JOYSTICK to compile
// joystick mode firmware. Joystick mode also implements the
// 'shift' function for 15 additional buttons. You can use
// the keymap setup mode to set player to other than 1 to
// disable the shift function.
// v0.8 7.12.2020
// Updated compiler to 5.097.
// Heavy optimizations in debouncing, autofire and data
// writes to minimize lag. New code looks more complicated
// but executes faster.


//#define J_ACE_JOYSTICK

#include "j_ace.h"

void main(void) {
	active_buttons.bytes.a=0;				// Clear active_buttons. Union can't be initialized
	active_buttons.bytes.b=0;				// in declaration.
	active_buttons.bytes.c=0;
	
	usb_init();
	port_e_pullups(TRUE);					// Activates weak pull-up on pin RA3
	port_b_pullups(0b11111111);				// Activates weak pull-up on all port B pins
	set_tris_c(0x3c);						// Set bits 0,1,6,7 as LED outputs. bit 2 is AUTOFIRE input
	
	output_high(PIN_LED_1);					// Turn LEDs off initially. Otherwise they are always on
	output_high(PIN_LED_2);					// in joystick mode
	output_high(PIN_LED_3);
	
	setup_timer_2(T2_DIV_BY_16,249,3);		// Setup timer2 to prescale 1:16, period 249(+1), postscale 3
											// So interrupt frequency is 12MHz / 16 / 250 / 3 = 1000 -> period 1 ms.
	load_jace_settings();					// Load settings from EEPROM (Autofire mode, keymap etc.)
	setup_keymap();							// Run keymap setup mode if button 1 is pressed during USB plug-in

	// Main loop, run it forever
	while (TRUE) {
		if (usb_enumerated()) {
			measure_millisec();
			//read_inputs();				// Reads inputs w/o debounce
			read_debounced_inputs();		// Reads and debounces inputs. Debounced data stored in active_buttons[] array
			handle_shift();					// Handles 'shift' function, which means that alternate keymap is used when start button is held down
			memcpy(output_buttons.bits,active_buttons.bits,sizeof(active_buttons.bits));	// Copy the active buttons to output array
			add_autofire();					// Adds autofire function. Includes also autofire setup mode.
			#ifdef	J_ACE_JOYSTICK
			usb_joystick_task();
			#else
			send_keycodes_trplkbd();		// Converts output_buttons[] array to keycodes and sends data via three HID keyboards
			#endif
			usb_task();						// It has been suggested that this function should be called periodically even if
											// not using polled USB. Might otherwise cause problems on USB3 ports.
		}
		usb_task();							// It has been suggested that this function should be called periodically even if
											// not using polled USB. Might otherwise cause problems on USB3 ports.
	}	// End of endless loop
}

void add_autofire() {	
	////////////////////////////////////////
	/// AUTOFIRE functions
	////////////////////////////////////////
	static int1 global_autofire = 0;
	static int1 autofire_high = 0;
	static int1 autofire_setup_mode = 0;
	static unsigned int16 autofire_activation_counter = 0;
	static unsigned int8 autofire_delay_counter = 0;

	if (millisec_passed) {
		// Autofire button press length detection counter:
		if (active_buttons.bits[16]) {
			autofire_activation_counter++;
		}
		else {
			autofire_activation_counter = 0;
		}
		
		// Normal autofire activation/deactivaion:
		if (autofire_activation_counter == 1)
			global_autofire = !global_autofire;
		
		// Autofire setup mode activation
		if (autofire_activation_counter >= 1024) {
			autofire_setup_mode = 1;
			global_autofire = 0;
		}	
		
		// Autofire speed counter
		autofire_delay_counter++;
		if (autofire_delay_counter >= autofire_speed_setting) {
			autofire_delay_counter = 0;
			autofire_high = !autofire_high;
		}	
		
		// Autofire indication LED
		if (global_autofire) {
			output_low(PIN_LED_4);
		}
		else {
			output_high(PIN_LED_4);					
		}
	}	
	
	// Autofire operation on activated buttons
	if (global_autofire) {
		if (active_buttons.bytes.a) {
			for (unsigned int8 i = 0; i<8; i++) {
				if (active_buttons.bits[i]) {
					if (autofire_buttons[i]) {
						if (previous_buttons[i]==0) {	// This ensures that button reacts immediately
							autofire_delay_counter=0;
							autofire_high=1;
						}	
						if (autofire_high) {
							output_buttons.bits[i] = 1;
							output_low(PIN_LED_4);
						}
						else {
							output_buttons.bits[i] = 0;
							output_high(PIN_LED_4);
						}			
					}
				}
			}
		}	
		if (active_buttons.bytes.b) {
			for (unsigned int8 i = 0; i<8; i++) {
				if (active_buttons.bits[i+8]) {
					if (autofire_buttons[i+8]) {
						if (previous_buttons[i+8]==0) {	// This ensures that button reacts immediately
							autofire_delay_counter=0;
							autofire_high=1;
						}	
						if (autofire_high) {
							output_buttons.bits[i+8] = 1;
							output_low(PIN_LED_4);
						}
						else {
							output_buttons.bits[i+8] = 0;
							output_high(PIN_LED_4);
						}			
					}
				}
			}
		}	
	}				

	// Autofire setup
	if (autofire_setup_mode) {
		if (autofire_high)				// indicates current autofire speed by autofire led
			output_low(PIN_LED_4);		// this also indicates that autofire setup mode is active
		else
			output_high(PIN_LED_4);
			
		if (autofire_activation_counter == 1) {
			autofire_setup_mode = 0;
			global_autofire = 0;
			memcpy(autofire_buttons,active_buttons.bits,sizeof(autofire_buttons));	// Copy the active buttons to autofire buttons array when exiting autofire setup
			write_eeprom(1,autofire_speed_setting);
			write_eeprom(2,*((unsigned int8 *)autofire_buttons));
			write_eeprom(3,*((unsigned int8 *)(autofire_buttons)+1));
		}
		else {
			if (active_buttons.bits[8] && !previous_buttons[8]) {	// joystick up -> increase autofire speed
				if (autofire_speed_setting > AUTOFIRE_DELAY_LOW_LIMIT) {
					autofire_speed_setting /= 2;
				}
			}
			if (active_buttons.bits[9] && !previous_buttons[9]) {	// joystick down -> decrease autofire speed
				if (autofire_speed_setting < AUTOFIRE_DELAY_HIGH_LIMIT) {
					autofire_speed_setting *= 2;
				}
			}
		}
	}
}

void send_keycodes_trplkbd() {
	unsigned int8 tx_msg[20];
	unsigned int8 leds;
	unsigned int8 butindex=2;
	unsigned int8 keycode_idx,modifier_idx;
	unsigned int8 i;
	unsigned int8 * buff_add;

	memset(tx_msg, 0x00, sizeof(tx_msg));
	
	// update tx_msg with keyboard scancodes for currently pressed down buttons
	keycode_idx = player+shift_active;
	modifier_idx = keycode_idx+1;
	
	if (output_buttons.bytes.a) {	
		for (int i=0;i<8;i++) {
			if (bit_test(output_buttons.bytes.a,i)) {
				if (kbd_scancodes[i][modifier_idx])					// The scancode for this button is a modifier key
					tx_msg[0] |= kbd_scancodes[i][keycode_idx];
				else											// The scancode for this button is a normal key
					tx_msg[butindex++] = kbd_scancodes[i][keycode_idx];
			}
		}
	}
	if (output_buttons.bytes.b) {	
		for (int i=0;i<8;i++) {
			if (bit_test(output_buttons.bytes.b,i)) {
				if (kbd_scancodes[i+8][modifier_idx])					// The scancode for this button is a modifier key
					tx_msg[0] |= kbd_scancodes[i+8][keycode_idx];
				else											// The scancode for this button is a normal key
					tx_msg[butindex++] = kbd_scancodes[i+8][keycode_idx];
			}
		}
	}	
	
	// For testing delay, sends Numlock keycode if autofire button is pressed
	// Oscilloscope can be used to measure time between button input falling edge
	// and LED1 falling edge. Currently (fw0.8) it is between 1.3 and 3.3 ms.
	//if (active_buttons.bits[16])
	//	tx_msg[butindex++] = KEY_NUMLOCK;

	// Send keycodes for 1st keyboard
	disable_interrupts(INT_USB);
	buff_add = EP_BDxADR_I(1);
   	memcpy(buff_add, tx_msg, 8);
	enable_interrupts(INT_USB);

	if (usb_tbe(1)) {
		EP_BDxCNT_I(1)=8;
        i = EP_BDxST_I(1);
         if (bit_test(i,6))
            EP_BDxST_I(1) = 0x88;  //was DATA1, goto DATA0
         else
            EP_BDxST_I(1) = 0xc8;  //was DATA0, goto DATA1
 	}     


   //receive NUM LOCK, CAPS LOCK, etc LED status from PC.
   if (usb_kbhit(1)) {
      usb_get_packet(1, &leds, 1);
      leds = (leds & 0x03) | ((leds & 0x0c)<<4);
      output_c(leds ^ 0xff);		// invert bits, led must be connected between +5V and LED output pin
   }
   
   // Send the remaining buttons (if any)
    memset(tx_msg+6, 0x00, 2);
	
	disable_interrupts(INT_USB);
	buff_add = EP_BDxADR_I(2);
   	memcpy(buff_add, tx_msg+6, 8);
	enable_interrupts(INT_USB);

	if (usb_tbe(2)) {
		EP_BDxCNT_I(2)=8;
        i = EP_BDxST_I(2);
         if (bit_test(i,6))
            EP_BDxST_I(2) = 0x88;  //was DATA1, goto DATA0
         else
            EP_BDxST_I(2) = 0xc8;  //was DATA0, goto DATA1
 	}     

    memset(tx_msg+12, 0x00, 2);
   	
   	disable_interrupts(INT_USB);
	buff_add = EP_BDxADR_I(3);
   	memcpy(buff_add, tx_msg+12, 8);
	enable_interrupts(INT_USB);

	if (usb_tbe(3)) {
		EP_BDxCNT_I(3)=8;
        i = EP_BDxST_I(3);
         if (bit_test(i,6))
            EP_BDxST_I(3) = 0x88;  //was DATA1, goto DATA0
         else
            EP_BDxST_I(3) = 0xc8;  //was DATA0, goto DATA1
 	}     
}

void handle_shift() {
	// Handle 'shift' function if configured as player 1 keyboard
	static int1 shift_used = 0;
	static int1 shift_allowed = 0;
	static unsigned int16 shift_counter = 0;
	shift_active = 0;

	if (player == 0) {
		if ((current_startbutton == 0) && (previous_buttons[12] == 1) && (shift_used == 0))	{	// Activate start button when it is released
			shift_counter = 100;
		}
		if ((current_startbutton == 1) && (previous_buttons[12] == 0)) {		// When start button is pressed,
			shift_used = 0;														// start monitorin if it is used as shift
		}
		if ((current_startbutton == 1) && (previous_buttons[12] == 0)) {		// Shift is allowed only if no other keys are being
			if((*(unsigned int16*)active_buttons.bits) & 0xefff)				// pressed when start button is pressed
				shift_allowed = 0;												// otherwise it wouldn't be possible to e.g.
			else																// continue game in 1942 because it requires
				shift_allowed = 1;												// fire being held down while pressing start
		}	
		if (current_startbutton == 1 && shift_allowed) {						// When start button is kept down, activate shift function
			active_buttons.bits[12] = 0;										// and clear the original active_buttons start button bit
			shift_active = 8;
		}
		if ((current_startbutton == 1) && *(unsigned int16*)active_buttons.bits && shift_allowed) {	// If start button is used as shift, do not send
			shift_used = 1;														// normal start button scancode!
		}
		if (shift_counter) {													// Keep start buttons pressed some time after release
			active_buttons.bits[12] = 1;
			if (millisec_passed)
				shift_counter--;
		}
	}
}


void read_debounced_inputs() {
	unsigned int8 input_shadow_a,input_shadow_b;
	unsigned int1 input_shadow_c;
	unsigned int8 *debounce_pt = debounce_timer;

	input_shadow_a = input_a();
	input_shadow_b = input_b();
	input_shadow_c = input(PIN_AUTOFIRE);
	
	active_buttons.bits[12] = current_startbutton;
	memcpy(previous_buttons,active_buttons.bits,sizeof(active_buttons.bits));
			
	// Debounce port A inputs
	for (int i=0;i<8;i++) {
		if (bit_test(input_shadow_a,i) && *debounce_pt==0) {						// Button not pressed, debouncer idle
		}
		else if (!bit_test(input_shadow_a,i) && *debounce_pt==PRESS_DEBOUNCE_MS) {	// Button held down, debouncer idle
		}
		else if (*debounce_pt!=0 && *debounce_pt < PRESS_DEBOUNCE_MS) {				// Press debounce in process
			if (millisec_passed)
				(*debounce_pt)++;
		}		
		else if (*debounce_pt > PRESS_DEBOUNCE_MS) {								// Release debounce in process
			if (millisec_passed)
				(*debounce_pt)++;
			if (*debounce_pt == TOTAL_DEBOUNCE_MS)
				*debounce_pt = 0;
		}		
		else if (!bit_test(input_shadow_a,i) && *debounce_pt==0) {					// Button just pressed, begin debounce
			(*debounce_pt)++;
			active_buttons.bits[i] = 1;				// 7 asm instructions
		}		
		else if (bit_test(input_shadow_a, i) && *debounce_pt == PRESS_DEBOUNCE_MS) {	// Button just released, begin debounce
			(*debounce_pt)++;															// Is the if really needed here(?)
			active_buttons.bits[i] = 0;
		}
		
		debounce_pt++;
	}

	// Debounce port B inputs
	for (int i=0;i<8;i++) {
		if (bit_test(input_shadow_b,i) && *debounce_pt==0) {						// Button not pressed, debouncer idle
		}
		else if (!bit_test(input_shadow_b,i) && *debounce_pt==PRESS_DEBOUNCE_MS) {	// Button held down, debouncer idle
		}
		else if (*debounce_pt!=0 && *debounce_pt < PRESS_DEBOUNCE_MS) {				// Press debounce in process
			if (millisec_passed)
				(*debounce_pt)++;
		}		
		else if (*debounce_pt > PRESS_DEBOUNCE_MS) {								// Release debounce in process
			if (millisec_passed)
				(*debounce_pt)++;
			if (*debounce_pt == TOTAL_DEBOUNCE_MS)
				*debounce_pt = 0;
		}		
		else if (!bit_test(input_shadow_b,i) && *debounce_pt==0) {					// Button just pressed, begin debounce
			(*debounce_pt)++;
			active_buttons.bits[i+8] = 1;				// 7 asm instructions
		}		
		else if (bit_test(input_shadow_b, i) && *debounce_pt == PRESS_DEBOUNCE_MS) {	// Button just released, begin debounce
			(*debounce_pt)++;															// Is the if really needed here(?)
			active_buttons.bits[i+8] = 0;
		}
		
		debounce_pt++;
	}

	// Debounce port C input (autofire)
	if (input_shadow_c && *debounce_pt==0) {						// Button not pressed, debouncer idle
	}
	else if (!input_shadow_c && *debounce_pt==PRESS_DEBOUNCE_MS) {	// Button held down, debouncer idle
	}
	else if (*debounce_pt!=0 && *debounce_pt < PRESS_DEBOUNCE_MS) {				// Press debounce in process
		if (millisec_passed)
			(*debounce_pt)++;
	}		
	else if (*debounce_pt > PRESS_DEBOUNCE_MS) {								// Release debounce in process
		if (millisec_passed)
			(*debounce_pt)++;
		if (*debounce_pt == TOTAL_DEBOUNCE_MS)
			*debounce_pt = 0;
	}		
	else if (!input_shadow_c && *debounce_pt==0) {					// Button just pressed, begin debounce
		(*debounce_pt)++;
		active_buttons.bits[16] = 1;
	}		
	else if (input_shadow_c && *debounce_pt == PRESS_DEBOUNCE_MS) {	// Button just released, begin debounce
		(*debounce_pt)++;															// Is the if really needed here(?)
		active_buttons.bits[16] = 0;
	}

	current_startbutton = active_buttons.bits[12];
}


void read_inputs() {
	unsigned int8 input_shadow_a,input_shadow_b;
	unsigned int1 input_shadow_c;

	input_shadow_a = input_a();
	input_shadow_b = input_b();
	input_shadow_c = input(PIN_AUTOFIRE);
	
	active_buttons.bits[12] = current_startbutton;
	memcpy(previous_buttons,active_buttons.bits,sizeof(active_buttons.bits));
	
	active_buttons.bytes.a = ~input_shadow_a;
	active_buttons.bytes.b = ~input_shadow_b;
	active_buttons.bits[16] = !input_shadow_c;
	
	current_startbutton = active_buttons.bits[12];
}


void load_jace_settings() {
	// Load J-ACE settings from EEPROM
	player = read_eeprom(0);
	autofire_speed_setting = read_eeprom(1);
	*((unsigned int8 *)autofire_buttons) = read_eeprom(2);
	*((unsigned int8 *)(autofire_buttons)+1) = read_eeprom(3);
	
	// Check validity of EEPROM data
	if (player>6)
		player=0;
	if (autofire_speed_setting > AUTOFIRE_DELAY_HIGH_LIMIT)
		autofire_speed_setting = AUTOFIRE_DEFAULT_DELAY;
	if (autofire_speed_setting < AUTOFIRE_DELAY_LOW_LIMIT)
		autofire_speed_setting = AUTOFIRE_DEFAULT_DELAY;
	if (autofire_buttons[8] && autofire_buttons[9]) // Joystick up and down
		*((unsigned int8 *)(autofire_buttons)+1) = 0x00;	
}

void setup_keymap() {
		// Setup player / keymap
	int i=0;
	if (!input(PIN_BUT_1)) {				// Player (keymap) selection entered by keeping button 1 pressed when plugging in USB
		while(!input(PIN_BUT_1)) {}			// wait release of button 1
		delay_ms(RELEASE_DEBOUNCE_MS);		// deboune the release
		while(input(PIN_BUT_1)) {			// wait again press of button 1
			i++;
			delay_ms(1);
			if (i<127)						// indicate setup mode by flashing leds
				output_c(0x00);
			else
				output_c(0xff);
		}
		if (!input(PIN_JOY_UP))
			player = 0;
		if (!input(PIN_JOY_RIGHT))
			player = 2;
		if (!input(PIN_JOY_DOWN))
			player = 4;
		if (!input(PIN_JOY_LEFT))
			player = 6;
		write_eeprom(0, player);
	}				
}

void usb_joystick_task(void) 
{
	unsigned int8 out_data[6],i;
	unsigned int8 * buff_add;

    memset(out_data, 0x00, sizeof(out_data));

	if (shift_active) {
		out_data[1] = *((unsigned int8 *)output_buttons.bits+1) << 4;					// Joystick directions as buttons 13..16
		out_data[2] = *((unsigned int8 *)output_buttons.bits);							// Buttons 1..8 as 17..24		
		out_data[2] = out_data[2]&0x3f | out_data[2]>>1&0x40 | out_data[2]<<1&0x80;		// Swap buttons 7 and 8
		out_data[3] = *((unsigned int8 *)output_buttons.bits+1) >> 5 & 0x07;			// coin, a and b as buttons 25..27
	}
	else {
		out_data[0] = *((unsigned int8 *)output_buttons.bits);
		out_data[0] = out_data[0]&0x3f | out_data[0]>>1&0x40 | out_data[0]<<1&0x80;		// Swap buttons 7 and 8
		out_data[1] = *((unsigned int8 *)output_buttons.bits+1) >> 4;					// Start, coin, A and B

		if(output_buttons.bits[IDX_JOY_UP])				// up
			out_data[5]=-126;
		else if(output_buttons.bits[IDX_JOY_DOWN])		// down
			out_data[5]=126;
	
	    if(output_buttons.bits[IDX_JOY_LEFT])			// left
			out_data[4]=-126;
		else if(output_buttons.bits[IDX_JOY_RIGHT])		// right
			out_data[4]=126;
	}	

	disable_interrupts(INT_USB);
	buff_add = EP_BDxADR_I(1);
   	memcpy(buff_add, out_data, 6);
	enable_interrupts(INT_USB);

	if (usb_tbe(1)) {
		EP_BDxCNT_I(1)=6;
        i = EP_BDxST_I(1);
         if (bit_test(i,6))
            EP_BDxST_I(1) = 0x88;  //was DATA1, goto DATA0
         else
            EP_BDxST_I(1) = 0xc8;  //was DATA0, goto DATA1
 	}     
}

void measure_millisec() {
	if (interrupt_active(INT_TIMER2)) {
		millisec_passed = 1;
		clear_interrupt(INT_TIMER2);
	}
	else {
		millisec_passed = 0;
	}
}			
		